"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileQuery = exports.compileAST = exports.simplifyConditionSets = exports.mediaConditionToConditionSets = exports.mediaFeatureToConditionSets = exports.invertConditionSet = exports.notConditionSets = exports.mergeConditionSets = exports.andConditionSets = exports.convertToStandardUnit = exports.RANGE_FEATURES = exports.DISCRETE_FEATURES = void 0;
const media_query_parser_1 = require("media-query-parser");
exports.DISCRETE_FEATURES = {
    "any-hover": { none: 1, hover: 1 },
    "any-pointer": { none: 1, coarse: 1, fine: 1 },
    "color-gamut": { srgb: 1, p3: 1, rec2020: 1 },
    grid: { 0: 1, 1: 1 },
    hover: { none: 1, hover: 1 },
    "overflow-block": { none: 1, scroll: 1, paged: 1 },
    "overflow-inline": { none: 1, scroll: 1 },
    pointer: { none: 1, coarse: 1, fine: 1 },
    scan: { interlace: 1, progressive: 1 },
    update: { none: 1, slow: 1, fast: 1 },
};
exports.RANGE_FEATURES = {
    "aspect-ratio": {
        feature: "aspect-ratio",
        type: "ratio",
        bounds: [false, [0, 1], [Infinity, 1], false],
    },
    "device-aspect-ratio": {
        feature: "device-aspect-ratio",
        type: "ratio",
        bounds: [false, [0, 1], [Infinity, 1], false],
    },
    color: {
        feature: "color",
        type: "integer",
        bounds: [true, 0, Infinity, false],
    },
    "color-index": {
        feature: "color-index",
        type: "integer",
        bounds: [true, 0, Infinity, false],
    },
    monochrome: {
        feature: "monochrome",
        type: "integer",
        bounds: [true, 0, Infinity, false],
    },
    "device-height": {
        feature: "device-height",
        type: "length",
        bounds: [true, 0, Infinity, false],
    },
    "device-width": {
        feature: "device-width",
        type: "length",
        bounds: [true, 0, Infinity, false],
    },
    height: {
        feature: "height",
        type: "length",
        bounds: [true, 0, Infinity, false],
    },
    width: {
        feature: "width",
        type: "length",
        bounds: [true, 0, Infinity, false],
    },
    resolution: {
        feature: "resolution",
        type: "resolution",
        bounds: [true, 0, Infinity, false],
    },
};
const DEFAULT_UNIT_CONVERSIONS = {
    // = 100vw
    widthPx: 1920,
    // = 100vh
    heightPx: 1080,
    // used to determine vi and vb
    writingMode: "horizontal-tb",
    // also used for rem
    emPx: 16,
    // also used for rlh
    lhPx: 16,
    // font-specific sizes
    exPx: 8,
    chPx: 8,
    capPx: 11,
    icPx: 16,
};
const convertToStandardUnit = (token, unitConversions) => {
    if (token.type === "<number-token>") {
        return {
            type: "number",
            value: token.value,
        };
    }
    else if (token.type === "<dimension-token>") {
        let unitType;
        switch (token.unit) {
            case "s":
            case "ms":
                unitType = "time";
                break;
            case "hz":
            case "khz":
                unitType = "frequency";
                break;
            case "dpi":
            case "dpcm":
            case "dppx":
            case "x":
                unitType = "resolution";
                break;
            default:
                unitType = "length";
        }
        if (token.unit === "px") {
            return {
                type: "dimension",
                subtype: "length",
                px: token.value,
            };
        }
        else if (unitType === "time") {
            return {
                type: "dimension",
                subtype: "time",
                ms: token.unit === "s" ? Math.round(token.value * 1000) : token.value,
            };
        }
        else if (unitType === "frequency") {
            return {
                type: "dimension",
                subtype: "frequency",
                hz: token.unit === "khz" ? Math.round(token.value * 1000) : token.value,
            };
        }
        else if (unitType === "resolution") {
            let dppx = token.value;
            if (token.unit === "dpi") {
                dppx = parseFloat((token.value * 0.0104166667).toFixed(3));
            }
            else if (token.unit === "dpcm") {
                dppx = parseFloat((token.value * 0.0264583333).toFixed(3));
            }
            return {
                type: "dimension",
                subtype: "resolution",
                dppx,
            };
        }
        else {
            if (token.unit in unitConversions) {
                const factor = unitConversions[token.unit];
                return {
                    type: "dimension",
                    subtype: "length",
                    px: parseFloat((token.value * factor).toFixed(3)),
                };
            }
            else {
                return {
                    type: "ident",
                    value: "{invalid}",
                };
            }
        }
    }
    else if (token.type === "<ident-token>") {
        if (token.value === "infinite") {
            return {
                type: "infinite",
            };
        }
        else {
            return {
                type: "ident",
                value: token.value,
            };
        }
    }
    else {
        return {
            type: "ratio",
            numerator: token.numerator,
            denominator: token.denominator,
        };
    }
};
exports.convertToStandardUnit = convertToStandardUnit;
const andConditionSets = (a, b) => {
    const newConditionSets = [];
    for (const x of a) {
        for (const y of b) {
            const newConditionSet = (0, exports.mergeConditionSets)(x, y);
            if (Object.keys(newConditionSet).length > 0) {
                newConditionSets.push(newConditionSet);
            }
        }
    }
    return newConditionSets;
};
exports.andConditionSets = andConditionSets;
const mergeConditionSets = (setA, setB) => {
    const output = Object.assign({}, setA);
    for (const k in setB) {
        const key = k;
        if (key in output) {
            const A = output;
            const B = setB;
            // in both, merge
            if (key === "media-type") {
                A[key] = B[key];
            }
            else if (key === "invalid-features") {
                A[key].push(...B[key]);
            }
            else if (output[key] === "{invalid}" || setB[key] === "{invalid}") {
                A[key] = "{invalid}";
            }
            else if (output[key] === "{never}" || setB[key] === "{never}") {
                A[key] = "{never}";
            }
            else {
                const A = output;
                const B = setB;
                if (key === "any-hover" ||
                    key === "any-pointer" ||
                    key === "overflow-block" ||
                    key === "overflow-inline" ||
                    key === "hover" ||
                    key === "grid" ||
                    key === "scan" ||
                    key === "pointer" ||
                    key === "update") {
                    // this is type safe, but ts can't work it out
                    A[key] = A[key] === B[key] ? A[key] : "{never}";
                }
                else if (key === "color-gamut") {
                    A[key] = [
                        A[key][0] && B[key][0],
                        A[key][1] && B[key][1],
                        A[key][2] && B[key][2],
                        A[key][3] && B[key][3],
                    ];
                }
                else {
                    const [aMinInclusive, aMin, aMax, aMaxInclusive] = A[key];
                    const aMinValue = typeof aMin === "number" ? aMin : aMin[0] / aMin[1];
                    const aMaxValue = typeof aMax === "number" ? aMax : aMax[0] / aMax[1];
                    const [bMinInclusive, bMin, bMax, bMaxInclusive] = B[key];
                    const bMinValue = typeof bMin === "number" ? bMin : bMin[0] / bMin[1];
                    const bMaxValue = typeof bMax === "number" ? bMax : bMax[0] / bMax[1];
                    let aMinMoreThanB = aMinInclusive === bMinInclusive ? false : !aMinInclusive;
                    if (aMinValue !== bMinValue)
                        aMinMoreThanB = aMinValue > bMinValue;
                    let aMaxLessThanB = aMaxInclusive === bMaxInclusive ? false : !aMaxInclusive;
                    if (aMaxValue !== bMaxValue)
                        aMaxLessThanB = aMaxValue < bMaxValue;
                    const mergedMinInclusive = aMinMoreThanB
                        ? aMinInclusive
                        : bMinInclusive;
                    const mergedMin = aMinMoreThanB ? aMin : bMin;
                    const mergedMax = aMaxLessThanB ? aMax : bMax;
                    const mergedMaxInclusive = aMaxLessThanB
                        ? aMaxInclusive
                        : bMaxInclusive;
                    // this is type safe, but ts can't work it out
                    A[key] = [
                        mergedMinInclusive,
                        mergedMin,
                        mergedMax,
                        mergedMaxInclusive,
                    ];
                }
            }
        }
        else {
            // this is type safe, but ts can't work it out
            output[key] = setB[key];
        }
    }
    return output;
};
exports.mergeConditionSets = mergeConditionSets;
const notConditionSets = (conditionSets) => {
    // !(a || b) = !a && !b
    return conditionSets
        .map((conditionSet) => (0, exports.invertConditionSet)(conditionSet))
        .reduce((a, b) => (0, exports.andConditionSets)(a, b));
};
exports.notConditionSets = notConditionSets;
const invertConditionSet = (set) => {
    if (Object.keys(set).length === 0) {
        return [];
    }
    else {
        let outputSets = [{}];
        for (const k in set) {
            const key = k;
            if (key === "invalid-features") {
                outputSets = outputSets.map((prevSet) => (Object.assign(Object.assign({}, prevSet), { "invalid-features": set[key] })));
            }
            else if (key === "media-type") {
                // ignore
            }
            else if (set[key] === "{invalid}") {
                outputSets = outputSets.map((prevSet) => (Object.assign(Object.assign({}, prevSet), { [key]: "{invalid}" })));
            }
            else if (set[key] === "{never}") {
                // do not add key (effectively removing it from result set)
            }
            else if (key in exports.DISCRETE_FEATURES) {
                const dKey = key;
                const dSet = set;
                if (dKey === "color-gamut") {
                    const prevGamutRange = dSet["color-gamut"];
                    outputSets = outputSets.map((set) => (Object.assign(Object.assign({}, set), { "color-gamut": [
                            !prevGamutRange[0],
                            !prevGamutRange[1],
                            !prevGamutRange[2],
                            !prevGamutRange[3],
                        ] })));
                }
                else if (dKey === "grid") {
                    const before = dSet[dKey];
                    outputSets = outputSets.map((prevSet) => (Object.assign(Object.assign({}, prevSet), { grid: before === 0 ? 1 : 0 })));
                }
                else {
                    const values = Object.keys(exports.DISCRETE_FEATURES[dKey]);
                    const before = dSet[dKey];
                    outputSets = outputSets.flatMap((prevSet) => values.reduce((sets, next) => {
                        if (next !== before) {
                            sets.push(Object.assign(Object.assign({}, prevSet), { [dKey]: next }));
                        }
                        return sets;
                    }, []));
                }
            }
            else if (key in exports.RANGE_FEATURES) {
                const rKey = key;
                const rSet = set;
                const [minInclusive, min, max, maxInclusive] = rSet[rKey];
                const isMinBounded = min !== -Infinity || !minInclusive;
                const isMaxBounded = max !== Infinity || !maxInclusive;
                if (isMinBounded && isMaxBounded) {
                    outputSets = outputSets.flatMap((set) => [
                        Object.assign(Object.assign({}, set), { [rKey]: [true, -Infinity, min, !minInclusive] }),
                        Object.assign(Object.assign({}, set), { [rKey]: [!maxInclusive, max, Infinity, true] }),
                    ]);
                }
                else if (!isMinBounded && !isMaxBounded) {
                    outputSets = outputSets.map((set) => (Object.assign(Object.assign({}, set), { [rKey]: "{never}" })));
                }
                else {
                    if (isMinBounded) {
                        outputSets = outputSets.map((set) => (Object.assign(Object.assign({}, set), { [rKey]: [true, -Infinity, min, !minInclusive] })));
                    }
                    else {
                        outputSets = outputSets.map((set) => (Object.assign(Object.assign({}, set), { [rKey]: [!maxInclusive, max, Infinity, true] })));
                    }
                }
            }
        }
        return outputSets;
    }
};
exports.invertConditionSet = invertConditionSet;
const mediaFeatureToConditionSets = (mediaFeature, unitConversions) => {
    let minValue = null;
    let minInclusive = true;
    let maxValue = null;
    let maxInclusive = true;
    if (mediaFeature.context === "range") {
        if (mediaFeature.range.leftToken !== null) {
            const { leftToken, leftOp } = mediaFeature.range;
            const value = (0, exports.convertToStandardUnit)(leftToken, unitConversions);
            if (leftOp === "<" || leftOp === "<=") {
                minValue = value;
                minInclusive = leftOp === "<=";
            }
            else if (leftOp === ">" || leftOp === ">=") {
                maxValue = value;
                maxInclusive = leftOp === ">=";
            }
            else {
                minValue = value;
                maxValue = value;
                minInclusive = true;
                maxInclusive = true;
            }
        }
        if (mediaFeature.range.rightToken !== null) {
            const { rightToken, rightOp } = mediaFeature.range;
            const value = (0, exports.convertToStandardUnit)(rightToken, unitConversions);
            if (rightOp === "<" || rightOp === "<=") {
                maxValue = value;
                maxInclusive = rightOp === "<=";
            }
            else if (rightOp === ">" || rightOp === ">=") {
                minValue = value;
                minInclusive = rightOp === ">=";
            }
            else {
                minValue = value;
                maxValue = value;
                minInclusive = true;
                maxInclusive = true;
            }
        }
    }
    else if (mediaFeature.context === "value") {
        const value = (0, exports.convertToStandardUnit)(mediaFeature.value, unitConversions);
        if (mediaFeature.prefix === "min") {
            minValue = value;
            minInclusive = true;
        }
        else if (mediaFeature.prefix === "max") {
            maxValue = value;
            maxInclusive = true;
        }
        else {
            minValue = value;
            maxValue = value;
            minInclusive = true;
            maxInclusive = true;
        }
    }
    if (!(mediaFeature.feature in exports.DISCRETE_FEATURES) &&
        !(mediaFeature.feature in exports.RANGE_FEATURES) &&
        mediaFeature.feature !== "orientation") {
        return mediaFeature.feature;
    }
    if (mediaFeature.context === "boolean") {
        const feature = mediaFeature.feature;
        if (mediaFeature.feature === "orientation") {
            return [{}];
        }
        else if (feature === "color-gamut") {
            // colorGamut has no 'none' value, but treats srgb as truthy
            return [
                {
                    "color-gamut": [false, true, true, true],
                },
            ];
        }
        else if (feature === "grid") {
            return [
                {
                    grid: 1,
                },
            ];
        }
        else if (feature === "any-hover" ||
            feature === "any-pointer" ||
            feature === "hover" ||
            feature === "overflow-block" ||
            feature === "overflow-inline" ||
            feature === "pointer" ||
            feature === "scan" ||
            feature === "update") {
            return (0, exports.invertConditionSet)({
                [feature]: "none",
            });
        }
        else if (feature === "aspect-ratio" ||
            feature === "device-aspect-ratio") {
            return [
                {
                    [feature]: [false, [0, 1], [Infinity, 1], true],
                },
            ];
        }
        else {
            /* feature === "color" ||
               feature === "color-index" ||
               feature === "monochrome" ||
               feature === "device-height" ||
               feature === "device-width" ||
               feature === "height" ||
               feature === "width" */
            return [
                {
                    [feature]: [false, 0, Infinity, true],
                },
            ];
        }
    }
    else if (mediaFeature.feature in exports.DISCRETE_FEATURES) {
        const feature = mediaFeature.feature;
        if (minValue === null || maxValue === null || minValue !== maxValue) {
            return [
                {
                    [feature]: "{invalid}",
                },
            ];
        }
        const unit = minValue;
        if (unit.type === "infinite" ||
            unit.type === "dimension" ||
            unit.type === "ratio") {
            return [
                {
                    [feature]: "{invalid}",
                },
            ];
        }
        else if (unit.type === "number") {
            if (feature !== "grid" || (unit.value !== 0 && unit.value !== 1)) {
                return [{ grid: "{invalid}" }];
            }
            else {
                return [
                    {
                        grid: unit.value,
                    },
                ];
            }
        }
        else {
            // unit.type === "ident"
            if (feature === "color-gamut") {
                const index = ["srgb", "p3", "rec2020"].indexOf(unit.value);
                if (index === -1) {
                    return [
                        {
                            "color-gamut": "{invalid}",
                        },
                    ];
                }
                else {
                    return [
                        {
                            "color-gamut": [false, index <= 0, index <= 1, index <= 2],
                        },
                    ];
                }
            }
            else {
                return [
                    {
                        [feature]: unit.value in exports.DISCRETE_FEATURES[feature]
                            ? unit.value
                            : "{invalid}",
                    },
                ];
            }
        }
    }
    else if (mediaFeature.feature === "orientation") {
        if (mediaFeature.context !== "range" &&
            mediaFeature.value.type === "<ident-token>") {
            const { value } = mediaFeature.value;
            if (value === "landscape") {
                return [
                    {
                        "aspect-ratio": [true, [1, 1], [Infinity, 1], false],
                    },
                ];
            }
            else if (value === "portrait") {
                return [
                    {
                        "aspect-ratio": [false, [0, 1], [1, 1], true],
                    },
                ];
            }
        }
        return [
            {
                "invalid-features": ["orientation"],
            },
        ];
    }
    else {
        // mediaFeature.feature in RANGE_FEATURES
        const featureData = exports.RANGE_FEATURES[mediaFeature.feature];
        if (featureData.type === "resolution") {
            let safeMinValue = null;
            let safeMaxValue = null;
            if (minValue !== null) {
                if (minValue.type === "dimension" &&
                    minValue.subtype === "resolution") {
                    safeMinValue = minValue.dppx;
                }
                else if (minValue.type === "infinite") {
                    safeMinValue = Infinity;
                }
                else {
                    safeMinValue = false;
                }
            }
            if (maxValue !== null) {
                if (maxValue.type === "dimension" &&
                    maxValue.subtype === "resolution") {
                    safeMaxValue = maxValue.dppx;
                }
                else if (maxValue.type === "infinite") {
                    safeMaxValue = Infinity;
                }
                else {
                    safeMaxValue = false;
                }
            }
            if (safeMinValue === false || safeMaxValue === false) {
                return [
                    {
                        resolution: "{invalid}",
                    },
                ];
            }
            else if (safeMinValue === null) {
                return [
                    {
                        resolution: [
                            true,
                            -Infinity,
                            safeMaxValue,
                            maxInclusive,
                        ],
                    },
                ];
            }
            else if (safeMaxValue === null) {
                return [
                    {
                        resolution: [
                            minInclusive,
                            safeMinValue,
                            Infinity,
                            true,
                        ],
                    },
                ];
            }
            else {
                if (safeMinValue > safeMaxValue ||
                    (safeMinValue === safeMaxValue && !(minInclusive && maxInclusive))) {
                    return [
                        {
                            resolution: "{never}",
                        },
                    ];
                }
                else {
                    return [
                        {
                            resolution: [
                                minInclusive,
                                safeMinValue,
                                safeMaxValue,
                                maxInclusive,
                            ],
                        },
                    ];
                }
            }
        }
        else if (featureData.type === "ratio") {
            const { feature } = featureData;
            let safeMinValue = null;
            let safeMaxValue = null;
            if (minValue !== null) {
                if (minValue.type === "ratio") {
                    safeMinValue = [minValue.numerator, minValue.denominator];
                }
                else if (minValue.type === "number" && minValue.value > 0) {
                    safeMinValue = [minValue.value, 1];
                }
                else {
                    safeMinValue = false;
                }
            }
            if (maxValue !== null) {
                if (maxValue.type === "ratio") {
                    safeMaxValue = [maxValue.numerator, maxValue.denominator];
                }
                else if (maxValue.type === "number" && maxValue.value > 0) {
                    safeMaxValue = [maxValue.value, 1];
                }
                else {
                    safeMaxValue = false;
                }
            }
            if (safeMinValue === false || safeMaxValue === false) {
                return [{ [feature]: "{invalid}" }];
            }
            else if (safeMinValue === null) {
                return [
                    {
                        [feature]: [
                            true,
                            [-Infinity, 1],
                            safeMaxValue,
                            maxInclusive,
                        ],
                    },
                ];
            }
            else if (safeMaxValue === null) {
                return [
                    {
                        [feature]: [
                            minInclusive,
                            safeMinValue,
                            [Infinity, 1],
                            true,
                        ],
                    },
                ];
            }
            else {
                if (safeMinValue > safeMaxValue ||
                    (safeMinValue === safeMaxValue && !(minInclusive && maxInclusive))) {
                    return [
                        {
                            resolution: "{never}",
                        },
                    ];
                }
                else {
                    return [
                        {
                            [feature]: [
                                minInclusive,
                                safeMinValue,
                                safeMaxValue,
                                maxInclusive,
                            ],
                        },
                    ];
                }
            }
        }
        else if (featureData.type === "integer") {
            const { feature } = featureData;
            let safeMinValue = null;
            let safeMaxValue = null;
            if (minValue !== null) {
                safeMinValue = minValue.type === "number" ? minValue.value : false;
            }
            if (maxValue !== null) {
                safeMaxValue = maxValue.type === "number" ? maxValue.value : false;
            }
            if (safeMinValue === false ||
                safeMaxValue === false ||
                !Number.isInteger(safeMinValue !== null && safeMinValue !== void 0 ? safeMinValue : 0) ||
                !Number.isInteger(safeMaxValue !== null && safeMaxValue !== void 0 ? safeMaxValue : 0)) {
                return [{ [feature]: "{invalid}" }];
            }
            else {
                if (safeMinValue === null) {
                    return [
                        {
                            [feature]: [
                                true,
                                -Infinity,
                                safeMaxValue,
                                maxInclusive,
                            ],
                        },
                    ];
                }
                else if (safeMaxValue === null) {
                    return [
                        {
                            [feature]: [
                                minInclusive,
                                safeMinValue,
                                Infinity,
                                true,
                            ],
                        },
                    ];
                }
                else {
                    if (safeMinValue > safeMaxValue ||
                        (safeMinValue === safeMaxValue && !(minInclusive && maxInclusive))) {
                        return [
                            {
                                resolution: "{never}",
                            },
                        ];
                    }
                    else {
                        return [
                            {
                                [feature]: [
                                    minInclusive,
                                    safeMinValue,
                                    safeMaxValue,
                                    maxInclusive,
                                ],
                            },
                        ];
                    }
                }
            }
        }
        else {
            // featureData.type === "length"
            const { feature } = featureData;
            let safeMinValue = null;
            let safeMaxValue = null;
            if (minValue !== null) {
                if (minValue.type === "number") {
                    safeMinValue = minValue.value === 0 ? 0 : false;
                }
                else {
                    safeMinValue =
                        minValue.type === "dimension" && minValue.subtype === "length"
                            ? minValue.px
                            : false;
                }
            }
            if (maxValue !== null) {
                if (maxValue.type === "number") {
                    safeMaxValue = maxValue.value === 0 ? 0 : false;
                }
                else {
                    safeMaxValue =
                        maxValue.type === "dimension" && maxValue.subtype === "length"
                            ? maxValue.px
                            : false;
                }
            }
            if (safeMinValue === false || safeMaxValue === false) {
                return [{ [feature]: "{invalid}" }];
            }
            else if (safeMinValue === null) {
                return [
                    {
                        [feature]: [
                            true,
                            -Infinity,
                            safeMaxValue,
                            maxInclusive,
                        ],
                    },
                ];
            }
            else if (safeMaxValue === null) {
                return [
                    {
                        [feature]: [
                            minInclusive,
                            safeMinValue,
                            Infinity,
                            true,
                        ],
                    },
                ];
            }
            else {
                if (safeMinValue > safeMaxValue ||
                    (safeMinValue === safeMaxValue && !(minInclusive && maxInclusive))) {
                    return [
                        {
                            [feature]: "{never}",
                        },
                    ];
                }
                else {
                    return [
                        {
                            [feature]: [
                                minInclusive,
                                safeMinValue,
                                safeMaxValue,
                                maxInclusive,
                            ],
                        },
                    ];
                }
            }
        }
    }
};
exports.mediaFeatureToConditionSets = mediaFeatureToConditionSets;
const mediaConditionToConditionSets = (mediaCondition, unitConversions) => {
    const conditionSetsSets = [];
    for (const child of mediaCondition.children) {
        if ("context" in child) {
            const result = (0, exports.mediaFeatureToConditionSets)(child, unitConversions);
            if (typeof result === "string") {
                conditionSetsSets.push([
                    {
                        "invalid-features": [result],
                    },
                ]);
            }
            else {
                conditionSetsSets.push(result);
            }
        }
        else {
            conditionSetsSets.push((0, exports.mediaConditionToConditionSets)(child, unitConversions));
        }
    }
    if (mediaCondition.operator === "or" || mediaCondition.operator === null) {
        return conditionSetsSets.flat();
    }
    else if (mediaCondition.operator === "and") {
        return conditionSetsSets.reduce((a, b) => (0, exports.andConditionSets)(a, b));
    }
    else {
        // "not"
        return (0, exports.notConditionSets)(conditionSetsSets[0]);
    }
};
exports.mediaConditionToConditionSets = mediaConditionToConditionSets;
const simplifyConditionSets = (conditionSets) => {
    const permutations = [];
    const invalidFeatures = new Set();
    const neverFeatures = new Set();
    for (const conditionSet of conditionSets) {
        let isUnmatchable = false;
        if (Array.isArray(conditionSet["invalid-features"]) &&
            conditionSet["invalid-features"].length > 0) {
            for (const invalidFeature of conditionSet["invalid-features"]) {
                invalidFeatures.add(invalidFeature);
            }
            isUnmatchable = true;
        }
        const permutation = {};
        for (const k in conditionSet) {
            const set = conditionSet;
            const key = k;
            if (key === "invalid-features") {
                continue;
            }
            else {
                if (key === "color-gamut") {
                    const prev = set[key].toString();
                    if (prev === "false,false,false,false") {
                        set[key] = "{never}";
                    }
                    else if (prev === "true,true,true,true") {
                        continue;
                    }
                }
                const value = set[key];
                if (value === "{invalid}") {
                    invalidFeatures.add(key);
                    isUnmatchable = true;
                }
                else if (value === "{never}") {
                    neverFeatures.add(key);
                    isUnmatchable = true;
                }
                else {
                    if (key !== "media-type" || value !== "all") {
                        if (key in exports.RANGE_FEATURES) {
                            const [minInclusive, min, max, maxInclusive] = set[key];
                            const minValue = typeof min === "number" ? min : min[0] / min[1];
                            const maxValue = typeof max === "number" ? max : max[0] / max[1];
                            const isMinLTEMax = minValue < maxValue ||
                                (minValue === maxValue && (!maxInclusive || minInclusive));
                            if (isMinLTEMax) {
                                const [lbInclusive, lb, ub, ubInclusive] = exports.RANGE_FEATURES[key].bounds;
                                const lbValue = typeof lb === "number" ? lb : lb[0] / lb[1];
                                const ubValue = typeof ub === "number" ? ub : ub[0] / ub[1];
                                const isMinLTELowerBound = minValue < lbValue ||
                                    (minValue === lbValue && (!lbInclusive || minInclusive));
                                const isMinGTUpperBound = minValue > ubValue ||
                                    (minValue === ubValue && (!ubInclusive || !minInclusive));
                                const isMaxLTLowerBound = maxValue < lbValue ||
                                    (maxValue === lbValue && (!lbInclusive || !maxInclusive));
                                const isMaxGTEUpperBound = maxValue > ubValue ||
                                    (maxValue === ubValue && (!ubInclusive || maxInclusive));
                                if (isMinGTUpperBound || isMaxLTLowerBound) {
                                    neverFeatures.add(key);
                                    isUnmatchable = true;
                                }
                                else if (isMinLTELowerBound && isMaxGTEUpperBound) {
                                    // {always}
                                }
                                else if (isMinLTELowerBound) {
                                    permutation[key] = [
                                        lbInclusive,
                                        lb,
                                        max,
                                        maxInclusive,
                                    ];
                                }
                                else if (isMaxGTEUpperBound) {
                                    permutation[key] = [
                                        minInclusive,
                                        min,
                                        ub,
                                        ubInclusive,
                                    ];
                                }
                                else {
                                    permutation[key] = [
                                        minInclusive,
                                        min,
                                        max,
                                        maxInclusive,
                                    ];
                                }
                            }
                            else {
                                neverFeatures.add(key);
                                isUnmatchable = true;
                            }
                        }
                        else {
                            permutation[key] = value;
                        }
                    }
                }
            }
        }
        if (!isUnmatchable) {
            permutations.push(permutation);
        }
    }
    return {
        permutations,
        invalidFeatures: [...invalidFeatures].sort(),
        neverFeatures: [...neverFeatures].sort(),
    };
};
exports.simplifyConditionSets = simplifyConditionSets;
const generateUnitConversions = (units) => {
    // increasing emPx should also increase other units,
    // but any units passed in override these defaults
    let impliedUnits = {};
    if (typeof units.emPx === "number") {
        impliedUnits = {
            exPx: Math.round(units.emPx * 0.5),
            chPx: Math.round(units.emPx * 0.5),
            capPx: Math.round(units.emPx * 0.7),
            icPx: Math.round(units.emPx),
        };
    }
    const mergedUnits = Object.assign(Object.assign(Object.assign({}, DEFAULT_UNIT_CONVERSIONS), impliedUnits), units);
    const { widthPx, heightPx, writingMode, emPx: em, lhPx: lh, exPx: ex, chPx: ch, capPx: cap, icPx: ic, } = mergedUnits;
    const vw = widthPx / 100;
    const vh = heightPx / 100;
    const vmin = Math.min(vh, vw);
    const vmax = Math.max(vh, vw);
    const vi = writingMode === "horizontal-tb" ? vw : vh;
    const vb = writingMode === "horizontal-tb" ? vh : vw;
    return {
        em,
        rem: em,
        lh,
        rlh: lh,
        ex,
        ch,
        cap,
        ic,
        vw,
        vh,
        vmin,
        vmax,
        vi,
        vb,
        cm: 37.79527559,
        mm: 0.03779527559,
        in: 96,
        q: 0.009448818898,
        pc: 16,
        pt: 16,
    };
};
const compileAST = (ast, units = {}) => {
    const unitConversions = generateUnitConversions(units);
    const allConditions = [];
    for (const mediaQuery of ast) {
        const extraConditions = [];
        if (mediaQuery.mediaPrefix === "not") {
            if (mediaQuery.mediaType === "print") {
                extraConditions.push({
                    "media-type": "not-print",
                });
            }
            else if (mediaQuery.mediaType === "screen") {
                extraConditions.push({
                    "media-type": "not-screen",
                });
            }
            if (mediaQuery.mediaCondition !== null) {
                extraConditions.push(...(0, exports.notConditionSets)((0, exports.mediaConditionToConditionSets)(mediaQuery.mediaCondition, unitConversions)).map((conditionSet) => {
                    if (mediaQuery.mediaType === "all") {
                        return conditionSet;
                    }
                    else {
                        return Object.assign(Object.assign({}, conditionSet), { "media-type": mediaQuery.mediaType });
                    }
                }));
            }
        }
        else {
            const mediaType = mediaQuery.mediaType;
            if (mediaQuery.mediaCondition === null) {
                extraConditions.push({
                    "media-type": mediaType,
                });
            }
            else {
                extraConditions.push(...(0, exports.mediaConditionToConditionSets)(mediaQuery.mediaCondition, unitConversions).map((conditionSet) => (Object.assign(Object.assign({}, conditionSet), { "media-type": mediaType }))));
            }
        }
        if (extraConditions.every((condition) => { var _a; return ((_a = condition["invalid-features"]) !== null && _a !== void 0 ? _a : []).length === 0; })) {
            allConditions.push(...extraConditions);
        }
    }
    return (0, exports.simplifyConditionSets)(allConditions);
};
exports.compileAST = compileAST;
const compileQuery = (query, units = {}) => {
    const ast = (0, media_query_parser_1.toAST)(query);
    if (ast === null) {
        throw new Error("Query string was not lexed due to a syntax error");
    }
    return (0, exports.compileAST)(ast, units);
};
exports.compileQuery = compileQuery;
//# sourceMappingURL=compile.js.map